@@ -3,11 +3,12 @@ import glob
 import pathlib
 from functools import wraps
 import inspect
-from .pformat import *
+from parse import parse as parse_
+from pformat import *
 
-__all__ = ['Paths', 'Path', 'paths']
+__all__ = ['Paths', 'Path', 'tree']
 
-def paths(root, paths=None):
+def tree(root='.', paths=None, data=None):
     '''Build paths from a directory spec.
 
     Arguments:
@@ -19,10 +20,19 @@ def paths(root, paths=None):
     '''
     if isinstance(root, dict):
         root, paths = '.', root
-    data = {'root': root} if root else {}
+    return Paths(
+        {v: Path(*k) for k, v in get_keys({'{root}': {'': 'root', **paths}})},
+        dict(data or {}, root=root))
+
+
+def parse(pattern, s):
+    r = parse_(pattern, s)
+    return r and r.named
+
+
+class UnderspecifiedError(KeyError):
+    pass
 
-    return Paths({v: Path(*k) for k, v in get_keys({'{root}': paths})},
-                 data)
 
 class Paths(object):
     '''
@@ -59,10 +69,10 @@ class Paths(object):
         for path in self._paths.values():
             path.parent = self
 
-    @wraps(paths)
+    @wraps(tree)
     @classmethod
     def define(cls, *a, **kw):
-        return paths(*a, **kw)
+        return tree(*a, **kw)
 
 
     @property
@@ -71,7 +81,7 @@ class Paths(object):
 
     @property
     def copy(self):
-        return Paths({name: path.copy for name, path in self._paths.items()},
+        return Paths({name: path.copy for name, path in self.paths.items()},
                      dict(self.data))
 
     def add(self, root, paths):
@@ -84,10 +94,14 @@ class Paths(object):
         Returns:
             The initialized Paths object
         '''
-        if not isinstance(paths, Paths):
-            paths = Paths.define(None, {root: paths})
+        root = (root if isinstance(root, Path) else self.paths[root]).path_pattern
+        paths = paths if isinstance(paths, Paths) else tree(paths)
+
+        # add paths to
+        self.paths.update(**{
+            k: p.repath(p.format_only(root=root))
+            for k, p in paths.paths.items()})
 
-        self.paths.update(paths.paths)
         for path in paths.paths.values():
             path.parent = self
         return self
@@ -95,25 +109,21 @@ class Paths(object):
     def __repr__(self):
         return '<Paths data={} \n{}\n>'.format(self.data, '\n'.join([
             '\t{} : {}'.format(name, self[name].maybe_format())
-            for name in self._paths
+            for name in self.paths
         ]))
 
     def __iter__(self):
-        return iter(self._paths)
+        return iter(self.paths)
 
     def __getitem__(self, name):
-        return self._paths[name].copy
+        return self.paths[name]
 
     def __getattr__(self, name):
         try:
-            return self[name]
+            return self.paths[name]
         except KeyError:
             raise AttributeError()
 
-    def get(self, name, **kw):
-        '''Get a copy of a path object, optionally adding '''
-        return self[name].update(**kw)
-
     def parse(self, path, name):
         '''Parse data from a formatted string (reverse of string format)
 
@@ -126,11 +136,13 @@ class Paths(object):
     def translate(self, file, form, to, **kw):
         return self[to].specify(**self[form].parse(file, **kw))
 
-    def mkdirs(self, exist_ok=True):
-        '''Instantiate all fully specified directories'''
-        for name, path in self.format().items():
-            if path and not isinstance(path, Path):
-                os.makedirs(os.path.dirname(path), exist_ok=exist_ok)
+    def makedirs(self):
+        '''Instantiate all fully specified directories.'''
+        for path in self.paths.values():
+            try:
+                path.make(up=1)
+            except UnderspecifiedError:
+                pass
 
     def update(self, **kw):
         '''Update format data in place.'''
@@ -145,12 +157,13 @@ class Paths(object):
         '''Remove keys from paths dictionary'''
         p = self.copy
         for key in keys:
-            del p.data[key]
+            p.data.pop(key, None)
         return p
 
     @property
     def fully_specified(self):
-        return all(p.fully_specified for p in self._paths)
+        '''Are all paths fully specified?'''
+        return all(p.fully_specified for p in self.paths.values())
 
     def format(self, **kw):
         '''Return a dictionary where all fully specified paths are converted to strings
@@ -198,6 +211,7 @@ class Path(os.PathLike):
     # missing will convert to glob pattern and return all matching files
     assert isinstance(path.matching)
     '''
+    __FORBIDDEN_KEYS__ = ()
     def __init__(self, *path, data=None, parent=None):
         self._path = pathlib.Path(*path)
         self.data = {} if data is None else data
@@ -211,38 +225,48 @@ class Path(os.PathLike):
         return self.format()
 
     def __repr__(self):
-        return '<Path "{}" data={}>'.format(self.path, self.path_data)
+        return '<Path "{}" data={}>'.format(self.path_pattern, self.path_data)
 
     def __getattr__(self, name):
-        return getattr(self._path, name)
+        return getattr(self.path, name)
+
+    @property
+    def path_pattern(self):
+        '''The path as an unformatted string'''
+        return str(self._path)
 
     @property
     def path(self):
-        '''The path object as a string (unformatted)'''
-        return str(self._path)
+        '''Get the formatted path as a pathlib.Path object'''
+        return pathlib.Path(self.format())
+
+    @property
+    def s(self):
+        '''Convert to string'''
+        return str(self)
 
     @property
     def path_data(self):
         '''Both the path specific data and the paths group data'''
-        return {**self.parent.data, **self.data}
+        return {**self.parent.data, **self.data} if self.parent else self.data
 
     def update(self, **kw):
         '''Update specified data in place'''
-        self.data.update(**kw)
+        self.data.update(**{k: v for k, v in kw.items() if k not in self.__FORBIDDEN_KEYS__})
         return self
 
     def specify(self, **kw):
         '''Update specified data and return a new object'''
         return self.copy.update(**kw)
 
-    def unspecify(self, *keys):
+    def unspecify(self, *keys, parent=True):
         '''Remove keys from path dictionary'''
         p = self.copy
-        if p.parent:
+        if parent and p.parent:
             p.parent = p.parent.unspecify(*keys)
 
         for key in keys:
-            del p.data[key]
+            p.data.pop(key, None)
         return p
 
     @property
@@ -259,15 +283,36 @@ class Path(os.PathLike):
         '''Get a path without any attached data.'''
         return Path(self._path)
 
+    @property
+    def safe(self):
+        '''Make sure the path does not go above root.'''
+        return self.repath(os.path.normpath(os.sep + str(self._path)).lstrip(os.sep))
+
+    def repath(self, *f, data=None):
+        '''Make a copy with an entirely new path.'''
+        return Path(*f, data=dict(self.data, **(data or {})), parent=self.parent)
+
+    def join(self, *f):
+        '''Make a copy and append directories to the end.'''
+        return self.repath(self._path, *f)
+
+    j = join
+
     @property
     def copy(self):
-        return Path(self._path, data=dict(self.data), parent=self.parent)
+        '''Create a copy of the path object.'''
+        return self.join()
+
+    def up(self, n=1):
+        '''Create a copy of the path object up one directory.'''
+        return self.repath(
+            os.path.normpath(os.path.join(self._path, *(['..']*n))))
 
     def find_sibling(self, name):
         '''Find another path in the root tree.'''
         try:
             return self.parent[name]
-        except AttributeError as e:
+        except (AttributeError, TypeError) as e:
             raise AttributeError('No related paths are available.')
         except KeyError as e:
             raise KeyError('No related paths by that name are available.')
@@ -278,14 +323,16 @@ class Path(os.PathLike):
         Raises:
             KeyError if the format string is underspecified.
         '''
-        return self.path.format(**{**self.path_data, **kw})
+        try:
+            return self.path_pattern.format(**{**self.path_data, **kw})
+        except KeyError as e:
+            raise UnderspecifiedError(str(e))
 
     def parse(self, path, use_data=True):
         '''Extract variables from a compiled path'''
-        from parse import parse
-        pattern = self.partial_format() if use_data else self.path
-        r = parse(pattern, path)
-        if not r:
+        pattern = self.partial_format() if use_data else self.path_pattern
+        data = parse(pattern, path)
+        if not data:
             raise ValueError(inspect.cleandoc('''
                 Could not parse path using pattern.
                     path:{}
@@ -295,11 +342,10 @@ class Path(os.PathLike):
                 so any specified keys will be fixed. This is helpful to dodge ambiguous parsing
                 cases. To disable this pass `use_data=False` to parse.
                 '''.format(path, pattern)))
-        data = r.named
         return {**self.path_data, **data}
 
     def translate(self, path, to, **kw):
-        '''Translate the paths from '''
+        '''Translate the paths to another pattern'''
         return self.find_sibling(to).specify(**self.parse(path, **kw))
 
     def maybe_format(self, **kw):
@@ -312,19 +358,65 @@ class Path(os.PathLike):
 
     def partial_format(self, **kw):
         '''Format a field, leaving all unspecified fields to be filled later.'''
-        return pformat(self.path, **{**self.path_data, **kw})
+        f = pformat(self.path_pattern, **{**self.path_data, **kw})
+        return f
+
+    def format_only(self, **kw):
+        return pformat(self.path_pattern, **kw)
 
     @property
     def glob_pattern(self):
         '''Format a field, setting all unspecified fields as a wildcard (asterisk).'''
-        return gformat(self.path, **self.path_data)
+        return gformat(self.path_pattern, **self.path_data)
 
-    def glob(self):
+    def glob(self, *f):
         '''Find all matching files. unspecified fields are set as a wildcard (asterisk).'''
-        return sorted(glob.glob(self.glob_pattern))
+        return sglob(self.glob_pattern, *f)
 
-    def iglob(self):
-        return glob.iglob(self.glob_pattern)
+    def iglob(self, *f):
+        '''Find all matching files as a generator.'''
+        return glob.iglob(os.path.join(self.glob_pattern, *f))
+
+    def rglob(self, *f):
+        '''Find all matching files recursively as a generator.'''
+        return self.path.rglob(os.path.join(*(f or '*')))
+
+    def make(self, up=0):
+        '''Create this (or up a) directory.'''
+        os.makedirs(self.up(up), exist_ok=True)
+        return self
+
+    def touch(self, *a, **kw):
+        '''Touch this file - will recursively create parent directories.'''
+        self.make(up=1).path.touch(*a, **kw)
+        return self
+
+    def rm(self):
+        '''Remove this file or directory.'''
+        p = self.safe
+        p.rmdir() if self.is_dir() else os.remove(p.format()) if self.is_file() else None
+        return self
+
+    def rmglob(self, *f, include=False):
+        '''Recursively remove files matching join(*f). Set include=True, to
+        remove this node as well.'''
+        fs = sorted(self.safe.rglob(*f), key=lambda p: p.parts, reverse=True)
+        for fi in fs:
+            fi.rmdir() if fi.is_dir() else os.remove(fi)
+        if include:
+            self.rm()
+        return self
+
+    def write(self, x, mode='', **kw):
+        '''Write to file. Set mode='b' to write as bytes.'''
+        self.make(1)
+        b = 'b' in mode if mode else isinstance(x, (bytes, bytearray))
+        self.write_bytes(x, **kw) if b else self.write_text(x, **kw)
+        return self
+
+    def read(self, mode='', **kw):
+        '''Read file. Set mode='b' to read as bytes.'''
+        return self.read_bytes(**kw) if 'b' in mode else self.read_text(**kw)
 
 
 
@@ -347,3 +439,16 @@ def get_keys(data, keys=None):
                 yield keys_, val
         else:
             yield keys_, value
+
+example = tree({
+    'data': {
+        '{date}': {
+            '': 'date',
+            '{labels_set}.csv': 'csv',
+            'flac': {
+                '': 'flac_root',
+                '{name}.flac': 'flac',
+            }
+        }
+    }
+})
